"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Clause_1 = require("../../models/Clause");
var deref_1 = require("../../utils/deref");
var fragment_1 = require("./fragment");
var indexedFragmentStates = function (fragment) {
    var nextIndex = 0;
    var frontier = [fragment.head];
    var states = [];
    while (frontier.length > 0) {
        var state = frontier.pop();
        if (state.index === null) {
            state.index = nextIndex;
            nextIndex++;
            state.transitions.forEach(function (transition) {
                frontier.push(transition.target);
            });
            states.push(state);
        }
    }
    return states;
};
var evalFunctions = {};
function evalClause(clause) {
    clause = deref_1.default(clause);
    var evalFn;
    if (clause.type === null) {
        throw 'Clause has no type: ' + clause;
    }
    else if (!(clause.type in evalFunctions)) {
        evalFn = evalFunctions.PRED;
    }
    else {
        evalFn = evalFunctions[clause.type];
    }
    var r = evalFn(clause);
    return r;
}
var evalChildThen = function (wrapper) {
    return function evalChildThenWrapped(clause) {
        var childFrag = evalClause(clause.exprs[0]);
        return wrapper(childFrag);
    };
};
var evalChildrenThen = function (wrapper) {
    return function evalChildrenThenWrapped(clause) {
        var childFrags = clause.exprs.map(function (child) {
            var s = evalClause(child.expr);
            s.name = child.name;
            return s;
        });
        return wrapper(childFrags);
    };
};
['ROOT',
    'Z_OR_M',
    'O_OR_M',
    'Z_OR_O'].forEach(function (fragName) {
    evalFunctions[fragName] = evalChildThen(fragment_1.default[fragName]);
});
['OR',
    'CAT'].forEach(function (fragName) {
    evalFunctions[fragName] = evalChildrenThen(fragment_1.default[fragName]);
});
evalFunctions.PRED = function (x) {
    return fragment_1.default['PRED'](x);
};
function wrapRoot(expr) {
    return new Clause_1.default({
        type: 'ROOT',
        exprs: [expr],
        conformFn: null,
        generateFn: null,
        opts: {}
    });
}
function compile(expr) {
    var rootedExpr = wrapRoot(expr);
    var fragment = evalClause(rootedExpr);
    var states = indexedFragmentStates(fragment);
    var numStates = states.length;
    var nfaTransitions = {};
    var finalState;
    states.forEach(function (state) {
        if (state.transitions.length === 0) {
            finalState = state.index;
        }
        var outTrans = {};
        state.transitions.forEach(function (fragTrans) {
            outTrans[fragTrans.target.index] = fragTrans.clause;
        });
        nfaTransitions[state.index.toString()] = outTrans;
    });
    return {
        initialState: 0,
        numStates: numStates,
        finalState: finalState,
        transitions: nfaTransitions,
        expression: expr,
    };
}
exports.default = compile;
;
//# sourceMappingURL=compile.js.map