"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ClauseRef_1 = require("../models/ClauseRef");
var core_1 = require("../core");
var isClause_1 = require("../utils/isClause");
var isPred_1 = require("../utils/isPred");
var isBool_1 = require("../preds/isBool");
var walk_1 = require("../walk");
var resolve_1 = require("./resolve");
var coerceIntoClause_1 = require("../utils/coerceIntoClause");
var oPath = require("./simpleObjectPath");
var utils_1 = require("../utils");
var namespace_types_1 = require("../clauses/namespace.types");
var reg;
var _get = core_1.fclause({
    args: core_1.cat(utils_1.isNamespacePath),
    ret: utils_1.isClauseRef,
}).instrument(_getUnchecked);
function _getUnchecked(ref) {
    function getFn(prefix) {
        var path = reg;
        if (prefix) {
            path = prefix + ref;
        }
        else {
            path = ref;
        }
        var nObj = oPath.get(reg, _slashToDot(path));
        if (nObj) {
            return Object.assign(nObj['.expr'], nObj['.meta']);
        }
        else {
            return undefined;
        }
    }
    var sr = new ClauseRef_1.default({ ref: ref, getFn: getFn, conformFn: null });
    sr.conform = function clauseRefConform(x) {
        var ss = getFn();
        return walk_1.default(ss, x, { conform: true });
    };
    return sr;
}
function _slashToDot(p) {
    return p.replace(/^(.+)(\/)(.+)$/, '$1.$3').replace(/^\//, '');
}
// var PartialRefMapClause = shape({
//   req: {
//     'refDefs': [isNamespacePath, ExprOrPartialRefMapClause]
//   }
// });
function getNamespacePath(_a) {
    var nsPath = _a.nsPath;
    var retVal;
    var nameObj = _get(nsPath);
    retVal = nameObj;
    return retVal;
}
function setNamespacePath(_a) {
    var nsPath = _a.nsPath, expression = _a.expression;
    _processVal(nsPath, expression);
}
function _processVal(prefix, expression) {
    if (expression) {
        if (expression.clause || expression.pred) {
            var expr = expression.clause || expression.pred;
            _set(prefix, { '.expr': expr });
            return expr;
        }
        else {
            console.error(expression);
            throw '!';
        }
        // TODO
        // } else if ( val.partialRefMap ) {
        //   var { refDefs } = val.partialRefMap;
        //   for ( var k in refDefs ) {
        //     if ( refDefs.hasOwnProperty( k ) ) {
        //       var retVal = _processVal( refDefs[ k ] );
        //     }
        //   }
    }
    else {
        console.error(expression);
        throw '!';
    }
}
var NameObjClause = core_1.shape({
    req: { '.expr': core_1.or(isClause_1.default, isPred_1.default) }
});
var _set = core_1.fclause({
    args: core_1.cat(utils_1.isNamespacePath, NameObjClause),
    ret: isBool_1.default,
}).instrument(function _set(n, nObj) {
    _maybeInitRegistry();
    var existing = oPath.get(reg, _slashToDot(n));
    oPath.set(reg, _slashToDot(n), Object.assign({}, existing, nObj));
    return true;
});
var K = '___CLAUSEJS_REGISTRY';
function _maybeInitRegistry() {
    if (!reg) {
        clearRegistry();
    }
    return reg;
}
function clearRegistry() {
    reg = global[K] = {};
}
exports.clearRegistry = clearRegistry;
exports.setMeta = namespace_types_1.SetMetaFnClause.instrumentConformed(function setMeta(_a) {
    var _b = _a.source, namespacePath = _b.namespacePath, expression = _b.expression, metaObj = _a.metaObj, registry = _a.registry;
    if (!registry) {
        registry = reg;
    }
    if (namespacePath) {
        var nObj = oPath.get(registry, _slashToDot(namespacePath));
        var currMeta = nObj && nObj['.meta'];
        oPath.set(registry, _slashToDot(namespacePath), Object.assign({}, nObj, { '.meta': Object.assign({}, currMeta, metaObj) }));
    }
    else if (expression) {
        var clause = coerceIntoClause_1.default(expression);
        clause.meta = Object.assign(clause.meta, metaObj);
    }
});
exports.getMeta = namespace_types_1.GetMetaFnClause.instrumentConformed(function getMeta(_a) {
    var _b = _a.source, namespacePath = _b.namespacePath, expression = _b.expression, registry = _a.registry;
    if (!registry) {
        registry = reg;
    }
    if (namespacePath) {
        var nObj = oPath.get(registry, _slashToDot(namespacePath));
        var meta = nObj && nObj['.meta'];
        return meta;
    }
    else if (expression) {
        var clause = coerceIntoClause_1.default(expression);
        return clause.meta;
    }
});
function resolve(expr, reg) {
    if (!reg) {
        return resolve_1.default(expr, exports.getRegistry());
    }
    else {
        return resolve_1.default(expr, reg);
    }
}
exports.resolve = resolve;
_maybeInitRegistry();
exports.getRegistry = function () { return reg; };
var namespaceGetOrSet = namespace_types_1.NamespaceFnClause.instrumentConformed(function namespaceGetOrSet(_a) {
    var register = _a.register, retrieve = _a.retrieve;
    if (register) {
        return setNamespacePath(register);
    }
    else if (retrieve) {
        return getNamespacePath(retrieve);
    }
});
exports.get = namespace_types_1.GetNSFnClause.instrumentConformed(getNamespacePath);
exports.set = namespace_types_1.SetNSFnClause.instrumentConformed(setNamespacePath);
namespaceGetOrSet.get = exports.get;
namespaceGetOrSet.set = exports.set;
namespaceGetOrSet.clearRegistry = clearRegistry;
namespaceGetOrSet.getRegistry = exports.getRegistry;
namespaceGetOrSet.setMeta = exports.setMeta;
exports.default = namespaceGetOrSet;
//# sourceMappingURL=index.js.map