"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fnName_1 = require("./fnName");
var clauseFromAlts_1 = require("./clauseFromAlts");
exports.SEPARATOR = '___CLAUSEJS_SEPARATOR___';
function LevelIn(lvl) {
    this.level = lvl;
    this.isLevelIn = true;
}
exports.LevelIn = LevelIn;
function LevelOut(lvl) {
    this.level = lvl;
    this.isLevelOut = true;
}
exports.LevelOut = LevelOut;
var multipleArgFragmenter = function (_a) {
    var named = _a.opts.named, exprs = _a.exprs;
    if (exprs.length === 0) {
        //empty case
        return [];
    }
    else if (named) {
        return exprs.reduce(function (curr, _a, idx) {
            var name = _a.name, expr = _a.expr;
            return curr
                .concat(["\"" + name + "\"", ', ', expr,])
                .concat(idx < exprs.length - 1 ? [', ', exports.SEPARATOR] : []);
        }, []);
    }
    else {
        return exprs.reduce(function (curr, _a, idx) {
            var expr = _a.expr;
            return curr
                .concat([expr])
                .concat(idx < exprs.length - 1 ? [', ', exports.SEPARATOR] : []);
        }, []);
    }
};
var singleArgFragmenter = function (_a) {
    var enclosedClause = _a.opts.enclosedClause;
    return [enclosedClause];
};
var Fragmenters = {
    'PRED': function (_a) {
        var predicate = _a.opts.predicate;
        return [predicate];
    },
    'WALL': function (_a) {
        var enclosedClause = _a.opts.enclosedClause;
        return [enclosedClause];
    },
    // TODO
    'AND': function (_a) {
        var conformedExprs = _a.opts.conformedExprs;
        return interpose(conformedExprs.map(clauseFromAlts_1.default), [', ', exports.SEPARATOR]);
    },
    'CAT': multipleArgFragmenter,
    'OR': multipleArgFragmenter,
    'Z_OR_M': singleArgFragmenter,
    'O_OR_M': singleArgFragmenter,
    'Z_OR_O': singleArgFragmenter,
    'COLL_OF': singleArgFragmenter,
    'ANY': function () { return []; },
    // TODO
    'MAP_OF': function () { return []; },
    // TODO
    'SHAPE': function (_a) {
        var _b = _a.opts.conformedArgs.shapeArgs, _c = _b.optionalFields, _d = _c === void 0 ? {} : _c, _e = _d.opt, opt = _e === void 0 ? null : _e, _f = _d.optional, optional = _f === void 0 ? null : _f, _g = _b.requiredFields, _h = _g === void 0 ? {} : _g, _j = _h.req, req = _j === void 0 ? null : _j, _k = _h.required, required = _k === void 0 ? null : _k;
        return ['{ ', exports.SEPARATOR, new LevelIn(1)]
            .concat(_trimTrailingSeparator([].concat((req || required) ? ['required: '].concat(_fieldDefToFrags((req || required))).concat([', ', exports.SEPARATOR]) : [])
            .concat((opt || optional) ? ['optional: '].concat(_fieldDefToFrags((opt || optional))).concat([', ', exports.SEPARATOR]) : [])))
            .concat([new LevelOut(1), ' }']);
    },
    'FCLAUSE': function (_a) {
        var _b = _a.opts, args = _b.args, ret = _b.ret, fn = _b.fn;
        return ['{ ', exports.SEPARATOR, new LevelIn(1)].concat(_trimTrailingSeparator([].concat(args ? ['args: ', args, ', ', exports.SEPARATOR] : [])
            .concat(ret ? ['ret: ', ret, ', ', exports.SEPARATOR] : [])
            .concat(fn ? ['fn: ', fnName_1.default(fn), '()', ', ', exports.SEPARATOR] : [])))
            .concat([new LevelOut(1), exports.SEPARATOR, ' }']);
    },
};
function _fieldDefToFrags(_a) {
    var fields = _a.fieldDefs.fields;
    var r = ['{ ', exports.SEPARATOR, new LevelIn(1)];
    for (var key in fields) {
        if (fields.hasOwnProperty(key)) {
            r.push("\"" + key + "\"");
            r.push(': ');
            var _b = fields[key], keyValExprPair = _b.keyValExprPair, valExpressionOnly = _b.valExpressionOnly;
            if (keyValExprPair) {
                var keyExpression = keyValExprPair.keyExpression, valExpression = keyValExprPair.valExpression;
                r = r.concat(['[',])
                    .concat([
                    exports.SEPARATOR, '<keyExpression>: ', clauseFromAlts_1.default(keyExpression), ', ',
                    exports.SEPARATOR, '<valExpression>: ', clauseFromAlts_1.default(valExpression)
                ])
                    .concat([']']);
            }
            else if (valExpressionOnly) {
                r.push(clauseFromAlts_1.default(valExpressionOnly));
            }
            r.push(', ');
            r.push(exports.SEPARATOR);
        }
    }
    return r.concat([new LevelOut(1), exports.SEPARATOR, ' }']);
}
function _trimTrailingSeparator(frags) {
    if (frags.length > 0) {
        return frags.slice(0, frags.length - 2);
    }
    else {
        return frags;
    }
}
function getFragments(clause) {
    var fragmenter = Fragmenters[clause.type];
    if (!fragmenter) {
        console.error(clause);
        throw new Error("Unsupported clause type " + clause.type);
    }
    return fragmenter(clause);
}
function interpose(arr, interArr) {
    if (arr.length === 0) {
        return arr;
    }
    else {
        return arr.reduce(function (acc, curr, idx) {
            if (idx < arr.length - 1) {
                return acc.concat([curr]).concat(interArr);
            }
            else {
                return acc.concat([curr]);
            }
        }, []);
    }
}
exports.default = getFragments;
//# sourceMappingURL=fragmenter.js.map