"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var anyWalker_1 = require("./anyWalker");
var predWalker_1 = require("./predWalker");
var wallWalker_1 = require("./wallWalker");
var fclauseWalker_1 = require("./fclauseWalker");
var shapeWalker_1 = require("./shapeWalker");
var andWalker_1 = require("./andWalker");
var collOfWalker_1 = require("./collOfWalker");
var mapOfWalker_1 = require("./mapOfWalker");
var clauseRefWalker_1 = require("./clauseRefWalker");
var delayedClauseWalker_1 = require("./delayedClauseWalker");
var nfaWalker_1 = require("./nfaWalker");
var coerceIntoClause_1 = require("../utils/coerceIntoClause");
var isProblem_1 = require("../utils/isProblem");
function walk(clause, x, opts) {
    var phase = opts.phase;
    var walker = _getWalker(clause);
    if (!phase) {
        // 2-pass algorithm:
        // in Pass 1 we just need to know if x validates to clause, and if so, how
        var intermediate = walker.trailblaze(x, Object.assign({ phase: 'trailblaze' }, opts));
        if (isProblem_1.default(intermediate)) {
            return intermediate;
        }
        else {
            // in Pass 2 we return conformed and/or instrumented results
            return walker.reconstruct(intermediate, Object.assign({ phase: 'reconstruct' }, opts));
        }
    }
    else if (walker[phase]) {
        return walker[phase](x, opts);
    }
    else {
        throw '!';
    }
}
exports.default = walk;
var walkerMap = {
    'OR': nfaWalker_1.default,
    'CAT': nfaWalker_1.default,
    'COLL_OF': collOfWalker_1.default,
    'ANY': anyWalker_1.default,
    'Z_OR_M': nfaWalker_1.default,
    'O_OR_M': nfaWalker_1.default,
    'Z_OR_O': nfaWalker_1.default,
    'PRED': predWalker_1.default,
    'WALL': wallWalker_1.default,
    'SHAPE': shapeWalker_1.default,
    'AND': andWalker_1.default,
    'CLAUSE_REF': clauseRefWalker_1.default,
    'DELAYED': delayedClauseWalker_1.default,
    'FCLAUSE': fclauseWalker_1.default,
    'MAP_OF': mapOfWalker_1.default,
};
function _getWalker(expr) {
    var clause = coerceIntoClause_1.default(expr);
    var walker = walkerMap[clause.type];
    if (!walker) {
        throw 'unsupported type ' + clause.type;
    }
    var r = walker(clause, walk);
    return r;
}
//# sourceMappingURL=index.js.map